set rcsId {$Id: stooop.tcl,v 1.32 1995/12/23 21:57:46 jfontain Exp $}

# uncomment the following line if you want to use the dynamically loadable extension
#load libstooop.so.2.0 stooop

# rename proc{} before it is overloaded, ignore error in case of multiple inclusion of this file
catch {rename proc ::proc}

if {[llength [info commands new]]==0} {
# if dynamic extension not available

# initialize object id counter only once even if this file is sourced several times
if {![info exists _newId]} {
    set _newId 0
}

::proc new {classOrId args} {
    # create an object of specified class or copy an existing object

    global _newId _class

    # use local variable for identifier because new can be invoked recursively
    if {[catch {expr $classOrId}]} {
        # first argument is a class
        set _class([set id [incr _newId]]) $classOrId
        # invoke the constructor for the class with optional arguments
        eval $classOrId::$classOrId $id $args
    } else {
        # first argument is an object identifier
        # copy source object to new object of identical class
        [set _class([set id [incr _newId]]) $_class($classOrId)]::_copy $id $classOrId
    }
    # return a unique object identifier
    return $id
}

::proc delete {args} {
    # delete one or more objects
    global _class

    foreach id $args {
        _delete $_class($id) $id
        unset _class($id)
    }
}

::proc _delete {class id} {
    # delete object data starting at specified class layer and going up the base class hierarchy if any
    # invokes the destructor for the object class and unset all the object data members for the class
    # the destructor will in turn delete the base classes layers

    $class::~$class $id
    # delete all this object array members if any (assume that they were stored as $class($id,memberName))
    global $class
    foreach name [array names $class $id,*] {
        unset ${class}($name)
    }
}

::proc classof {id} {
    # return class of object
    global _class

    return $_class($id)
}

::proc _copy {class from to} {
    # copy object data members from one object to another

    global $class

    # copy regular data members
    set index [string length $from]
    foreach name [array names $class $from,*] {
        set ${class}($to[string range $name $index end]) [set ${class}($name)]
    }
    # copy array data members
    set index [string length $class$from]
    foreach name [info globals $class$from*] {
        global [set target $class$to[string range $name $index end]] $name
        array set $target [array get $name]
    }
}

# end of Tcl code used if dynamic extension not available
}

::proc virtual {keyword name arguments args} {
    # virtual operator, to be placed before proc
    # virtualize a member procedure or destructor, determine whether it is a pure virtual
    # check for procedures that cannot be virtualized

    if {[string compare $keyword proc]!=0} {
        error "virtual operator works only on proc, not $keyword"
    }
    if {![regexp {^(.+)::(.+)$} $name dummy class procedure]} {
        error "$name is not a valid member procedure name"
    }
    if {[string compare $class $procedure]==0} {
        error "cannot make a constructor virtual (class $class)"
    }
    if {[string compare ~$class $procedure]==0} {
        error "cannot make a destructor virtual (class $class)"
    }
    if {[string compare [lindex $arguments 0] this]!=0} {
        error "cannot make a static procedure ($name) virtual"
    }
    # set a global flag so that proc{} knows it is acting upon a virtual procedure, also serves as a pure indicator
    global _pureVirtual
    # no procedure body means pure virtual
    set _pureVirtual [expr [llength $args]==0]
    # process procedure declaration, body being empty for pure virtual procedure
    proc $name $arguments [lindex $args 0]
    unset _pureVirtual
}


::proc proc {name arguments args} {
    # preprocess class member procedure definitions by adding specific object oriented code
    # regular procedures are not affected
    # it is assumed that both constructor and destructor are defined for every class

    global _baseClasses

    if {![regexp {^(.+)::(.+)$} $name dummy class procedure]} {
        # not a class member procedure, fall back to normal procedure declaration
        ::proc $name $arguments [lindex $args 0]
        return
    }
    # check for procedure body presence
    if {[llength $args]==0} {
        error "missing body for $name"
    }
    if {[string compare $class $procedure]==0} {
        # class constructor definition
        if {[string compare [lindex $arguments 0] this]!=0} {
            error "class constructor first argument must be this (class $class)"
        }
        if {[catch {info body $class::$class}]} {
            # first constructor declaration: must be main constructor
            eval _constructorDeclaration $class 0 \{$arguments\} $args
            _generateDefaultCopyConstructor $class
        } else {
            # second constructor declaration: must be user defined copy constructor
            if {[llength $arguments]!=2} {
                error "class copy constructor must have 2 arguments exactly (class $class)"
            }
            if {[string compare [lindex $arguments 1] copy]!=0} {
                error "class copy constructor second argument must be copy (class $class)"
            }
            eval _constructorDeclaration $class 1 \{$arguments\} $args
        }
    } elseif {[string compare ~$class $procedure]==0} {
        # class destructor declaration
        if {[llength $arguments]!=1} {
            error "class destructor must have 1 argument exactly (class $class)"
        }
        if {[string compare [lindex $arguments 0] this]!=0} {
            error "class destructor argument must be this (class $class)"
        }
        # use fastest method for testing procedure existence
        if {[catch {info body $class::$class}]} {
            # make sure of proper declaration order
            error "class $class destructor defined before constructor"
        }
        _destructorDeclaration $class $arguments [lindex $args 0]
    } else {
        # regular member procedure, may be static if there is no this first argument
        _memberProcedureDeclaration $class $procedure $arguments [lindex $args 0]
    }
}

::proc _ancestors {class} {
    global _baseClasses

    if {![info exists _baseClasses($class)]} {
        return {}
    }
    foreach base [set ancestors $_baseClasses($class)] {
        foreach class [_ancestors $base] {
            if {[lsearch -exact $ancestors $class]<0} {
                # avoid duplicates
                lappend ancestors $class
            }
        }
    }
    return $ancestors
}

::proc _constructorDeclaration {class copy arguments args} {
    # copy flag is set for user defined copy constructor
    global _baseClasses

    # check that each base class constructor has arguments
    set number [expr [llength $args]-1]
    if {($number%2)!=0} {
        error "bad $class constructor declaration, a base class, contructor arguments or body may be missing"
    }

    # check base classes and their constructor arguments
    for {set index 0} {$index<$number} {incr index} {
        set base [lindex $args $index]
        if {!$copy} {
            # check and save base classes only for main constructor that defines them
            if {[info exists _baseClasses($class)]&&([lsearch -exact $_baseClasses($class) $base]>=0)} {
                error "class $class directly inherits from class $base more than once"
            }
            lappend _baseClasses($class) $base
        }
        # remove new lines in base arguments part in case user has formatted long declarations with new lines
        regsub -all \n [lindex $args [incr index]] {} constructorArguments($base)
    }
    set body [lindex $args $index]

    if {[info exists _baseClasses($class)]} {
        # base class(es) derivation specified
        foreach base $_baseClasses($class) {
            if {[string compare $class $base]==0} {
                error "class $class cannot be derived from itself"
            }
            if {[catch {info body $base::$base}]} {
                # make sure of proper declaration order
                error "class $class constructor defined before base class $base constructor"
            }
        }
        # setup access to this class and base classes data
        set constructorBody \
"
global [_ancestors $class] $class
"
        # set base part hidden derived member
        # invoke base class constructors before evaluating constructor body
        if {[string compare [lindex $arguments end] args]==0} {
            # variable number of arguments in derived class constructor
            foreach base $_baseClasses($class) {
                if {![info exists constructorArguments($base)]} {
                    error "missing base class $base constructor arguments from class $class constructor"
                }
                if {[string compare [lindex $constructorArguments($base) end] \$args]==0} {
                    # variable number of arguments in base class constructor
                    # use eval so that base class constructor sees arguments instead of a list
                    append constructorBody \
"set ${base}(\$this,_derived) $class
eval $base::$base \$this {[lrange $constructorArguments($base) 0 [expr [llength $constructorArguments($base)]-2]]} \$args
"
                } else {
                    # constant number of arguments in base class constructor
                    append constructorBody \
"set ${base}(\$this,_derived) $class
$base::$base \$this $constructorArguments($base)
"
                }
            }
        } else {
            # constant number of arguments
            foreach base $_baseClasses($class) {
                if {![info exists constructorArguments($base)]} {
                    error "missing base class $base constructor arguments from class $class constructor"
                }
                append constructorBody \
"set ${base}(\$this,_derived) $class
$base::$base \$this $constructorArguments($base)
"
            }
        }
    } else {
        # no base class derivation specified
        # setup access to this class data
        set constructorBody \
"
global $class
"
    }
    if {$copy} {
        # for user defined copy constructor, copy derived class member if it exists
        append constructorBody \
"catch {set ${class}(\$this,_derived) \[set ${class}(\$[lindex $arguments 1],_derived)\]}
"
    }
    append constructorBody $body
    if {$copy} {
        ::proc $class::_copy $arguments $constructorBody
    } else {
        ::proc $class::$class $arguments $constructorBody
    }
}

::proc _destructorDeclaration {class arguments body} {
    global _baseClasses

    # setup access to this class and base classes data
    # evaluate this class destructor body
    set body \
"
global [_ancestors $class] $class
$body
"
    if {[info exists _baseClasses($class)]} {
        # if there are any, delete base classes parts in reverse order of construction
        for {set index [expr [llength $_baseClasses($class)]-1]} {$index>=0} {incr index -1} {
            set base [lindex $_baseClasses($class) $index]
            append body \
"_delete $base \$this
"
        }
    }
    ::proc $class::~$class $arguments $body
}

::proc _memberProcedureDeclaration {class name arguments body} {
    global _pureVirtual

    if {[info exists _pureVirtual]} {
        # virtual declaration
        if {$_pureVirtual} {
            # pure virtual declaration
            # setup access to this class and base classes data
            # evaluate derived procedure which must exists. derived procedure return value is automatically returned
            ::proc $class::$name $arguments \
"
global [_ancestors $class] $class
eval \$${class}(\$this,_derived)::$name \[lrange \[info level 0\] 1 end\]
"
        } else {
            # regular virtual declaration
            # setup access to this class and base classes data
            # evaluate derived procedure and return if it exists
            # else evaluate the base class procedure which can be invoked from derived class procedure by prepending ::
            ::proc ::$class::$name $arguments \
"
global [_ancestors $class] $class
$body
"
            ::proc $class::$name $arguments \
"
global [_ancestors $class] $class
if {!\[catch {info body \$${class}(\$this,_derived)::$name}\]} {
return \[eval \$${class}(\$this,_derived)::$name \[lrange \[info level 0\] 1 end\]\]
}
eval ::\[info level 0\]
"
        }
    } else {
        # non virtual declaration
        # setup access to this class and base classes data
        ::proc $class::$name $arguments \
"
global [_ancestors $class] $class
$body
"
    }
}

::proc _baseCloningCode {class siblingArgumentName} {
    # generate code for cloning base classes layers

    global _baseClasses

    if {![info exists _baseClasses($class)]} {
        return {}
    }
    # there is at least one base class
    foreach base $_baseClasses($class) {
        append body \
"$base::_copy \$this \$$siblingArgumentName
"
    }
    return $body
}

::proc _generateDefaultCopyConstructor {class} {
    # generate default copy procedure which may be overriden by the user for any class layer

    set body [_baseCloningCode $class sibling]
    append body \
"_copy $class \$sibling \$this
"
    ::proc $class::_copy {this sibling} $body
}
